﻿
Imports System.Globalization
Imports Microsoft.Kinect
Imports System.IO
Imports System.Math
Imports System.EventArgs

Imports System.Threading

Imports System.Net

Module modKinect
    Public Class classKinect

        Dim timeAccelerometerDistanceUpdated As DateTime
        Dim boolFaceReplyInProgress As Boolean = False
        Private audioStream As Stream
        Dim boolKinectInitializationInProgress As Boolean = False
        Dim intScanVectors(128) As Integer

        Public Function ftnKinectInitializeTop() As Boolean
            On Error Resume Next
            ' Look through all sensors and start the first connected one.
            ftnKinectInitializeTop = False
            If Not boolKinectInitializationInProgress And Not boolGetKinectDataRunning And Not boolShutDown Then '
                boolKinectInitializationInProgress = True
                subReportSensorStatus()
                If (Not boolKinectConnected) Then
                    ftnKinectInitializeTop = ftnKinectInit()
                    subReportSensorStatus()
                Else
                    ftnKinectInitializeTop = True
                End If
            Else
                ftnKinectInitializeTop = True
            End If
            boolKinectInitializationInProgress = False
        End Function

        Function ftnKinectInit() As Boolean
            On Error Resume Next
            ftnKinectInit = False
            subConnectSensor()
            If boolKinectConnected Then
                ftnKinectInit = ftnKinectRun()
                subEnableDepthStream()
                tv4AccelerometerDistance.W = 0
                tv4AccelerometerDistance.X = 0
                tv4AccelerometerDistance.Y = 0
                tv4AccelerometerDistance.Z = 0

                tv4AccelerometerOffset = tv4AccelerometerAverage
                'System.Windows.Forms.Application.DoEvents()
                timeAccelerometerDistanceUpdated = Now()
            End If
        End Function

        Sub subConnectSensor()
            On Error Resume Next
            ftnKinectStop()
            For Each potentialSensor In KinectSensor.KinectSensors
                If potentialSensor.Status = KinectStatus.Connected Then
                    K_Sensor = potentialSensor
                    Exit For
                Else
                    K_Sensor = Nothing
                End If
            Next potentialSensor
            ftnKinectWaitInitializing()
            boolKinectConnected = (K_Sensor.Status = KinectStatus.Connected)
            If boolKinectConnected Then frmKinectConfigure.subUpdateStatusLabel1("Operating.", Color.Green)
        End Sub


        Sub subReportSensorStatus()
            On Error Resume Next
            boolKinectConnected = (K_Sensor.Status = KinectStatus.Connected)
            If (K_Sensor Is Nothing) Then
                frmKinectConfigure.subUpdateStatusLabel1("No available Kinect!", Color.Red)
                boolKinectConnected = False
            ElseIf (K_Sensor.Status = KinectStatus.Disconnected) Then
                frmKinectConfigure.subUpdateStatusLabel1("No available Kinect!", Color.Red)
                boolKinectConnected = False
            ElseIf (K_Sensor.Status = KinectStatus.NotPowered) Then
                frmKinectConfigure.subUpdateStatusLabel1("Kinect not powered!", Color.Red)
                boolKinectConnected = False
            ElseIf (K_Sensor.Status = KinectStatus.Error) Then
                frmKinectConfigure.subUpdateStatusLabel1("Kinect error detected!", Color.Red)
                boolKinectConnected = False
            ElseIf (K_Sensor.Status = KinectStatus.InsufficientBandwidth) Then
                frmKinectConfigure.subUpdateStatusLabel1("Kinect insufficient bandwidth!", Color.Red)
                boolKinectConnected = False
            ElseIf (K_Sensor.Status = KinectStatus.NotReady) Then
                frmKinectConfigure.subUpdateStatusLabel1("Kinect not ready!", Color.Red)
                boolKinectConnected = False
            ElseIf (Not K_Sensor.IsRunning) Then
                frmKinectConfigure.subUpdateStatusLabel1("Kinect Not running!", Color.Red)
                boolKinectConnected = False
            ElseIf Not boolKinectConnected Then
                frmKinectConfigure.subUpdateStatusLabel1("Kinect unknown error!", Color.Red)
            Else
                frmKinectConfigure.subUpdateStatusLabel1("Operating.", Color.Green)
            End If
        End Sub

        Function ftnKinectWaitInitializing() As Boolean
            On Error Resume Next
            ftnKinectWaitInitializing = False
            If K_Sensor IsNot Nothing Then
                Dim timeOut As DateTime = Now().AddMilliseconds(5000)
                Do Until (K_Sensor.Status <> KinectStatus.Initializing) Or boolShutDown Or (Now() > timeOut)
                    frmKinectConfigure.subUpdateStatusLabel1("Kinect Initializing!", Color.Red)
                    System.Windows.Forms.Application.DoEvents()
                    Thread.Sleep(50)
                Loop
                ftnKinectWaitInitializing = (K_Sensor.Status <> KinectStatus.Initializing)
            End If
        End Function


        Public Function ftnKinectRun() As Boolean
            On Error Resume Next
            ftnKinectRun = False
            If K_Sensor IsNot Nothing Then
                Dim timeOut As DateTime = Now().AddMilliseconds(5000)
                K_Sensor.ForceInfraredEmitterOff = False
                Do Until K_Sensor.IsRunning Or boolShutDown Or (Now() > timeOut)
                    K_Sensor.Start()
                    frmKinectConfigure.subUpdateStatusLabel1("Kinect Starting!", Color.Red)
                    System.Windows.Forms.Application.DoEvents()
                    Thread.Sleep(50)
                Loop
                ftnKinectRun = K_Sensor.IsRunning
            End If
        End Function


        Function ftnKinectStop() As Boolean
            On Error Resume Next
            ftnKinectStop = False
            If (K_Sensor IsNot Nothing) Then
                K_Sensor.ForceInfraredEmitterOff = True
                K_Sensor.ColorStream.Disable()
                K_Sensor.DepthStream.Disable()
                K_Sensor.AudioSource.Stop()
                ' RemoveHandler K_Sensor.AudioSource.BeamAngleChanged, AddressOf subAudioSourceBeamChanged
                ' RemoveHandler K_Sensor.AudioSource.SoundSourceAngleChanged, AddressOf subAudioSourceSoundSourceAngleChanged
                RemoveHandler K_Sensor.DepthFrameReady, AddressOf subGetKinectDepthData
                Dim timeOut As DateTime = Now().AddMilliseconds(5000)
                Do Until (Not K_Sensor.IsRunning) Or boolShutDown Or (Now() > timeOut) Or (K_Sensor Is Nothing)
                    K_Sensor.Stop()
                    frmKinectConfigure.subUpdateStatusLabel1("Stopping!", Color.Red)
                    System.Windows.Forms.Application.DoEvents()
                    Thread.Sleep(50)
                Loop
                subWait(1000)
                ftnKinectStop = Not K_Sensor.IsRunning
                If Not K_Sensor.IsRunning Then frmKinectConfigure.subUpdateStatusLabel1("Kinect Stoped!", Color.Red)
            End If
        End Function


        Public Sub subEnableDepthStream()
            On Error Resume Next
            If (K_Sensor IsNot Nothing) Then
                K_Sensor.ForceInfraredEmitterOff = False
                K_Sensor.ColorStream.Disable()
                AddHandler K_Sensor.DepthFrameReady, AddressOf subGetKinectDepthData
                K_Sensor.DepthStream.Enable(DepthImageFormat.Resolution640x480Fps30)
                K_Sensor.DepthStream.Range = DepthRange.Default
                boolGetKinectDataRunning = False
                K_Sensor.ElevationAngle() = intDefaultElevation
            End If
        End Sub

        Const intMiddleDepthSample As Integer = 640 * 320
        Dim boolGetKinectDataRunning As Boolean = False
        Public Sub subGetKinectDepthData(ByVal sender As Object, ByVal e As DepthImageFrameReadyEventArgs)
            Try
                If Not boolKinectInitializationInProgress And Not boolGetKinectDataRunning Then
                    boolGetKinectDataRunning = True
                    Dim int5TimesI As Short
                    Dim int127MinusI As Short
                    Using depthFrame As DepthImageFrame = e.OpenDepthImageFrame()
                        If depthFrame IsNot Nothing Then
                            Dim depthPixels(depthFrame.PixelDataLength - 1) As DepthImagePixel
                            depthFrame.CopyDepthImagePixelDataTo(depthPixels) ' Copy the pixel data from the image to a temporary array
                            For I As Short = 0 To 127
                                int5TimesI = 5 * I
                                int127MinusI = 127 - I
                                intScanVectors(int127MinusI) = depthPixels(Int(intMiddleDepthSample + (int5TimesI))).Depth
                                'If (intScanVectors(int127MinusI) = 0) Then intScanVectors(int127MinusI) = depthPixels(Int(intMiddleDepthSample + (int5TimesI + 1))).Depth
                            Next (I)
                            depthFrame.Dispose()
                        End If
                    End Using
                    'ftnBumpDetect()
                    'ftnAccelerometerDistance()
                    For intBeamNumber As Short = 0 To 127
                        trvScanVectors(intBeamNumber).sngRange = ftnCmToMapUnits(intScanVectors(intBeamNumber) * sngKinectCalCoefficients(intBeamNumber))
                    Next intBeamNumber
                    evntVectorsAreFresh.Set()
                    boolGetKinectDataRunning = False
                End If
            Catch ex As Exception
                frmDebug.subWriteLogs(Color.Green, " EXCEPTION in subGetKinectDepthData")
            End Try
        End Sub


        '------------------  Kinect Audio  -----------------------------------
        'Private Sub subAudioSourceSoundSourceAngleChanged(ByVal sender As Object, ByVal e As SoundSourceAngleChangedEventArgs)
        '    On Error Resume Next
        '    intAudioAngle = -e.Angle
        '    frmKinectConfigure.proSndDirection.Value = 50 + intAudioAngle
        'End Sub

        'Private Sub subAudioSourceBeamChanged(ByVal sender As Object, ByVal e As BeamAngleChangedEventArgs)
        '    'intAudioAngle = -e.Angle
        'End Sub




    End Class

End Module

