﻿

' This module returns a bot pose based upon Range Scan Data.
' It is used for devices which can return a 2D range scan such as a Kinect or Laser Range scanner
Imports System
Imports System.Net
Imports System.Net.Sockets
Imports System.Text
Imports System.Threading
Imports Microsoft.VisualBasic
Imports System.EventArgs
Imports System.Math
Imports Microsoft.Kinect

Module modLocalize2Dscan
    Public Const sngWeakMatch As Single = 0.5
    Public Const sngGoodMatch As Single = 0.35
    Public Const sngExcellentMatch As Single = 0.25
    Dim bool2DScanLocalizationInProgress As Boolean = False ' dont want conflict between modLocalizeTop and subLocalizeWhileMoving2Dscan_Top

    Public Function ftnLocalizeFrom2DScan_Top(ByRef tbpEstimatedPose As typeBotPose, ByVal tbpReferencePose As typeBotPose, ByVal boolBotIsMoving As Boolean, _
                                              ByRef boolGood2DLocalization As Boolean, ByRef sngRMS_Error As Single) As typeBotPose
        Try
            Dim tbpRawScanPose As typeBotPose
            sngRMS_Error = sngOneDotFiveSigma
            If boolBotIsMoving Or ftnIAmLost(tbpEstimatedPose) Then
                intNumberOfBeams = Max(16, intNumberOfBeamsDefault \ 2) ' reduce number of vectors while moving or lost for faster localization
            Else
                intNumberOfBeams = Max(16, intNumberOfBeamsDefault)
            End If
            boolGood2DLocalization = False
            boolAbortLocalization = False
            ftnLocalizeFrom2DScan_Top = tbpEstimatedPose
            If (boolKinectConnected Or boolRangersAreEnabled) And Not boolStop And Not boolSimulatedMovement Then
                Dim strStatusDrivingRoot As String = strStatusDriving
                Dim boolGoodPreVal2DLocalization As Boolean = False
                If Not boolBotIsMoving Then subDoEventsIfMainThread()
                If bool2DScanLocalizationInProgress Then
                    Dim timeOut As DateTime = Now().AddMilliseconds(12000)
                    While bool2DScanLocalizationInProgress And Not boolStop   ' wait for previous scan to finish
                        boolAbortLocalization = True
                        subDoEventsIfMainThread()
                        Thread.Sleep(25)
                        If Now() >= timeOut Then
                            ftnLocalizeFrom2DScan_Top = ftnLocalizeUsingDeadReckoning(ftnLocalizeFrom2DScan_Top, constGotAKinectDelay, tbpTargetPoseGlobal)
                            frmDebug.subWriteLogs(Color.Green, "  ftnLocalizeFrom2DScan_Top timeOut ")
                            Exit Function
                        End If
                    End While
                End If
                boolAbortLocalization = False
                bool2DScanLocalizationInProgress = True
                tbpRawScanPose = ftn2dLocalize(tbpEstimatedPose, boolGoodPreVal2DLocalization, sngRMS_Error)
                If Not boolGoodPreVal2DLocalization Then
                    ftnLocalizeFrom2DScan_Top = tbpEstimatedPose
                Else
                    If (sngRMS_Error < sngGoodMatch) Then
                        ftnLocalizeFrom2DScan_Top = tbpRawScanPose
                    ElseIf (sngRMS_Error < sngWeakMatch) Then
                        ftnLocalizeFrom2DScan_Top = ftnFindAvgPose(tbpEstimatedPose, tbpRawScanPose, 0)
                    End If
                    ftnLocalizeFrom2DScan_Top = ftnMinPoseUncertainty(ftnLocalizeFrom2DScan_Top)
                    If boolBotIsMoving Then tbpReferencePose = tbpEstimatedPose
                    ftnLocalizeFrom2DScan_Top = ftnValidatePose(ftnLocalizeFrom2DScan_Top, tbpReferencePose, boolGood2DLocalization)
                End If
                If Not boolBotIsMoving Then ' update graphics and log file with results
                    strStatusDriving = strStatusDrivingRoot & " 2D=" & FormatNumber(sngRMS_Error, 2) & "; "
                    GotABotClass.subRedraw(ftnLocalizeFrom2DScan_Top)
                    If boolGood2DLocalization Then subDrawScanVectors(tbpRawScanPose, True)
                    frmDebug.WritePose(ftnElapsedTime() & "       Pose Derived From 2D Scan Data  ", tbpRawScanPose)
                    If boolGood2DLocalization Then
                        frmDebug.subWriteLogs(Color.Green, "                         Scan Coefficient = " & FormatNumber(sngRMS_Error, 2))
                    Else
                        frmDebug.subWriteLogs(Color.Red, "                           Scan Coefficient = " & FormatNumber(sngRMS_Error, 2) & "   (localization failed)")
                    End If
                End If
            Else
            End If
            bool2DScanLocalizationInProgress = False
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Green, " EXCEPTION in ftnLocalizeFrom2DScan_Top")
        End Try
    End Function




    '--------------------------------------  Debug Graphics  ----------------------------------------------------------------------

    Public Sub subDrawScanVectors(ByVal tbpPose As typeBotPose, ByVal boolGreen As Boolean)
        Try
            If boolShowMore And Not boolBotIsMoving And Not GotABotClass.InvokeRequired Then
                Dim sngOrientaionOffset As Single = 29.25 + tbpPose.Orientation
                Dim sngOrientationFactor As Single = 58.5 / 128
                If boolKinectConnected Then
                    Dim tmcVectorOrigin As typeMapCoordinate = ftnLocationFromEstimatedPose(tbpPose, tmcScanVectorOffset)
                    Dim intStep As Short = Max(128 \ intNumberOfBeamsDefault, 2)
                    For intBeamNumber As Short = 0 To 127 Step intStep ' Draw Scan Vectors in Green or Red
                        If (trvScanVectors(intBeamNumber).sngRange > 0) Then _
                                subDrawVector(tmcVectorOrigin, trvScanVectors(intBeamNumber).sngRange, _
                                              ftnConvertTo360(sngOrientaionOffset - sngOrientationFactor * intBeamNumber), boolGreen) ' Show Scan Vector on map
                    Next
                End If
                If boolRangersAreEnabled And boolGreen Then ' Draw Ranger Vectors  
                    Dim trvRangerBeams(intHighestEnabledRangerNumber) As typeRangeVector  ' Range vector - relative to Bot center
                    For intRangerNumber As Short = 1 To intHighestEnabledRangerNumber ' 
                        If trfRangeFinders(intRangerNumber).boolEnable Then
                            trvRangerBeams(intRangerNumber) = frmRangeFinder.ftnRangeFinderVector(tbpPose, intRangerNumber, trfRangeFinders(intRangerNumber).sngRange, trfRangeFinders(intRangerNumber).intOrientation) ' Find Range vector - relative to Bot center
                            subDrawVector(tbpPose.Location, trvRangerBeams(intRangerNumber).sngRange, _
                                ftnConvertTo360(tbpPose.Orientation + trvRangerBeams(intRangerNumber).sngAngle), boolGreen)
                        End If
                    Next
                End If
                If boolGreen Then subWait(100)
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Green, " EXCEPTION in subDrawScanVectors")
        End Try
    End Sub

    Dim trvScanVectorsCopy(128) As typeRangeVector
    Dim boolftn2dLocalizeInProgress As Boolean = False
    Function ftn2dLocalize(ByVal tbpEstimatedPose As typeBotPose, _
                           ByRef boolGood2DLocalization As Boolean, ByRef sngRMS_Error As Single) As typeBotPose
        ftn2dLocalize = tbpEstimatedPose
        Try
            If Not boolftn2dLocalizeInProgress Then
                boolftn2dLocalizeInProgress = True
                sngRMS_Error = sngOneDotFiveSigma
                If Not boolBotIsMoving Then frmRangeFinder.subGetRangerReadings()
                If boolKinectConnected Or (boolRangersAreEnabled And Not boolBotIsMoving) Then
                    If boolKinectConnected Then
                        'If boolDebugKinect Then frmDebug.subWriteLogs(Color.Black, " pre aquire Kinect Vectors:  " & ftnElapsedTime() & vbCrLf)
                        If evntVectorsAreFresh.WaitOne(1000) Then
                            If boolDebugKinect Then frmDebug.subWriteLogs(Color.Black, "  Kinect Vectors acquired by ftn2dLocalize:  " & ftnElapsedTime() & vbCrLf)
                            trvScanVectorsCopy = trvScanVectors
                            subDrawScanVectors(tbpEstimatedPose, False)
                        Else
                            frmDebug.subWriteLogs(Color.Green, "  TimeOut waiting for Kinect vectors")
                            Exit Function
                        End If
                    End If
                    If boolKinectConnected Then _
                      If ftnNonFlatWall() Then ftn2dLocalize = ftnPoseFrom2DScan(tbpEstimatedPose, sngRMS_Error)
                End If
                'If ftnIAmLost(tbpEstimatedPose) Then
                '    boolGood2DLocalization = (sngRMS_Error <= sngWeakMatch)
                If (ftnDistanceBetweenLocations(tbpEstimatedPose.Location, ftn2dLocalize.Location) > sngFiveInches) Or _
                       (ftnAngleDifference(tbpEstimatedPose.Orientation, ftn2dLocalize.Orientation) > 6) Then
                    boolGood2DLocalization = (sngRMS_Error <= sngGoodMatch)
                Else
                    boolGood2DLocalization = (sngRMS_Error < sngWeakMatch)
                End If
            End If
            boolftn2dLocalizeInProgress = False
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Green, " EXCEPTION in ftn2dLocalize")
        End Try
    End Function




    Public Function ftnPoseFrom2DScan(ByVal tbpEstimatedPose As typeBotPose, ByRef sngRMS_Error As Single) As typeBotPose
        Try
            sngRMS_Error = 0
            ftnPoseFrom2DScan = ftnPoseWithLowestAvgRangeError(tbpEstimatedPose, sngRMS_Error, sngFourInches, 2)
            ftnPoseFrom2DScan.LocationUncertainty.X = sngFourInches / sngGridSize
            'If boolBotIsMoving Then
            'ftnPoseFrom2DScan.LocationUncertainty.X = sngFourInches / sngGridSize
            'Else
            '    ftnPoseFrom2DScan.LocationUncertainty.X = sngSixInches / sngGridSize
            'End If
            ftnPoseFrom2DScan.LocationUncertainty.Y = ftnPoseFrom2DScan.LocationUncertainty.X
            ftnPoseFrom2DScan.OrientationUncertainty = 4
            ftnPoseFrom2DScan = ftnPoseWithLowestAvgRangeError(ftnPoseFrom2DScan, sngRMS_Error, sngOneInch, 1)
        Catch ex As Exception
            ftnPoseFrom2DScan = tbpEstimatedPose
            frmDebug.subWriteLogs(Color.Green, " EXCEPTION in ftnPoseFrom2DScan")
        End Try
    End Function


    Dim boolAbortLocalization As Boolean = False
    Function ftnPoseWithLowestAvgRangeError(ByVal tbpEstimatedPose As typeBotPose, ByRef sngRMS_ErrorBest As Single, _
                                                 ByVal intLocationStep As Short, ByVal intOrientationStep As Short) As typeBotPose
        Try ' Find the pose within the bot's uncertainty region that most closely resembles the range data
            Dim intOrientationStart, intOrientationEnd As Short
            Dim tbpPossiblePose As typeBotPose = tbpEstimatedPose
            Dim intStartX, intEndX, intStartY, intEndY As Short
            Dim sngRMS_Error, sngRMS_ErrorEndPoint, sngRMS_ErrorEndPointBest As Single
            Dim boolSkipNextCheck As Boolean = False
            Dim intValidBeamCounter As Short = 0
            Dim intValidBeamCounterBest As Short = 0
            Dim sngPoseDifferenceFactor As Single
            ftnPoseWithLowestAvgRangeError = tbpEstimatedPose
            intStartX = tbpEstimatedPose.Location.X - tbpEstimatedPose.LocationUncertainty.X
            intEndX = tbpEstimatedPose.Location.X + tbpEstimatedPose.LocationUncertainty.X
            intStartY = tbpEstimatedPose.Location.Y - tbpEstimatedPose.LocationUncertainty.Y
            intEndY = tbpEstimatedPose.Location.Y + tbpEstimatedPose.LocationUncertainty.Y
            subCheckMapArrayLimits(intStartX, intEndX, intStartY, intEndY)
            subFindStartEndAngles(tbpEstimatedPose, intOrientationStart, intOrientationEnd)
            sngRMS_ErrorEndPointBest = ftnRMS_RangeError(tbpEstimatedPose, True, intValidBeamCounterBest) ' Find End Point Model error at estimated pose
            sngRMS_ErrorBest = ftnRMS_RangeError(tbpEstimatedPose, False, intValidBeamCounterBest) ' Find Ray Cast Model error at estimated pose
            If boolDebugLocalize2D And Not boolBotIsMoving Then
                frmDebug.subWriteLogs(Color.Green, " ================ Starting sngRMS_ErrorBest = " & FormatNumber(sngRMS_ErrorBest, 3) & "==============================")
                frmDebug.subWriteLogs(Color.Green, " Pose =  " & tbpEstimatedPose.Location.X & ",  " & tbpEstimatedPose.Location.Y & " at " & tbpEstimatedPose.Orientation & " degrees")
                frmDebug.subWriteLogs(Color.Green, " intLocationStep = " & intLocationStep)
                frmDebug.subWriteLogs(Color.Green, " intValidBeamCounter = " & intValidBeamCounterBest)
            End If
            For tbpPossiblePose.Location.X = intStartX To intEndX Step intLocationStep ' try to find a lower error pose 
                If boolStop Or boolAbortLocalization Then
                    ftnPoseWithLowestAvgRangeError = tbpEstimatedPose
                    sngRMS_ErrorBest = sngTwoSigma
                    Exit Function
                End If
                For tbpPossiblePose.Location.Y = intStartY To intEndY Step intLocationStep
                    If Not ftnInWall(tbpPossiblePose.Location) Then
                        For tbpPossiblePose.Orientation = intOrientationStart To intOrientationEnd Step intOrientationStep
                            tbpPossiblePose.Orientation = ftnConvertTo360(tbpPossiblePose.Orientation)
                            If boolSkipNextCheck Then
                                boolSkipNextCheck = False
                            Else
                                sngRMS_ErrorEndPoint = ftnRMS_RangeError(tbpPossiblePose, True, intValidBeamCounter) ' Find End Point Model error at this pose
                                If (sngRMS_ErrorEndPoint > Min(sngRMS_ErrorEndPointBest * 0.5, 1.5)) Then boolSkipNextCheck = True ' Skip the next check
                                If (sngRMS_ErrorEndPoint < sngRMS_ErrorEndPointBest) Then
                                    'If (ftnAngleDifference(tbpPossiblePose.Orientation, ftnPoseWithLowestAvgRangeError.Orientation) < 2) And _
                                    '   (ftnDistanceBetweenLocations(tbpPossiblePose.Location, ftnPoseWithLowestAvgRangeError.Location) <= sngThreeInches) Then  ' Don't need to check for intervening walls
                                    '    ftnPoseWithLowestAvgRangeError = tbpPossiblePose
                                    '    sngRMS_ErrorEndPointBest = sngRMS_ErrorEndPoint
                                    '    sngRMS_ErrorBest = sngRMS_ErrorEndPoint
                                    '    intValidBeamCounterBest = intValidBeamCounter
                                    'Else
                                    sngRMS_Error = ftnRMS_RangeError(tbpPossiblePose, False, intValidBeamCounter) ' Find Ray Cast Model error at this pose
                                    sngPoseDifferenceFactor = 1 + (ftnMapUnitsToInches(Abs(tbpEstimatedPose.Location.Y - tbpPossiblePose.Location.Y) + Abs(tbpEstimatedPose.Location.X - tbpPossiblePose.Location.X)) * sngGridSize + Abs(ftnAngleDifference(tbpEstimatedPose.Orientation, tbpPossiblePose.Orientation))) / 500
                                    If (sngRMS_Error * sngPoseDifferenceFactor < sngRMS_ErrorBest) Then ' Poses nearest estimated have larger weighting
                                        If boolDebugLocalize2D And Not boolBotIsMoving Then
                                            frmDebug.subWriteLogs(Color.Green, " ===================================================================")
                                            frmDebug.subWriteLogs(Color.Green, "  New sngRMS_ErrorBest = " & FormatNumber(sngRMS_Error, 3))
                                            frmDebug.subWriteLogs(Color.Green, " Pose =  " & tbpPossiblePose.Location.X & ",  " & tbpPossiblePose.Location.Y & " at " & tbpPossiblePose.Orientation & " degrees")
                                            frmDebug.subWriteLogs(Color.Green, " intLocationStep = " & intLocationStep)
                                            frmDebug.subWriteLogs(Color.Green, " intValidBeamCounter = " & intValidBeamCounter)
                                        End If
                                        ftnPoseWithLowestAvgRangeError = tbpPossiblePose
                                        sngRMS_ErrorEndPointBest = sngRMS_ErrorEndPoint
                                        sngRMS_ErrorBest = sngRMS_Error
                                        intValidBeamCounterBest = intValidBeamCounter
                                    End If
                                    'End If
                                End If
                            End If
                        Next tbpPossiblePose.Orientation
                    End If
                Next tbpPossiblePose.Location.Y
            Next tbpPossiblePose.Location.X
            sngRMS_ErrorBest = ftnMapUnitsToInches(sngRMS_ErrorBest)
            ftnPoseWithLowestAvgRangeError.Orientation = ftnConvertTo360(ftnPoseWithLowestAvgRangeError.Orientation)
            ftnPoseWithLowestAvgRangeError.OrientationUncertainty = tbpEstimatedPose.OrientationUncertainty
            ftnPoseWithLowestAvgRangeError.LocationUncertainty = tbpEstimatedPose.LocationUncertainty
            ftnPoseWithLowestAvgRangeError = ftnPoseUncertaintyFrom2DScan(ftnPoseWithLowestAvgRangeError, sngRMS_ErrorBest)
        Catch ex As Exception
            ftnPoseWithLowestAvgRangeError = tbpEstimatedPose
            frmDebug.subWriteLogs(Color.Green, " EXCEPTION in ftnPoseWithLowestAvgRangeError")
        End Try
    End Function

    Function ftnRMS_RangeError(ByVal tbpPossiblePose As typeBotPose, _
                                      ByVal boolEndPointModel As Boolean, ByRef intValidBeamCounter As Short) As Single
        ' boolEndPointModel = Don't check for intervening walls
        Try
            Dim intRangerValidBeams, intNumberOfRangers As Short
            ftnRMS_RangeError = ftnScanToMapRMS_Error(tbpPossiblePose, intValidBeamCounter, boolEndPointModel)
            If Not boolBotIsMoving And boolRangersAreEnabled Then
                Dim trvRangerBeams(intHighestEnabledRangerNumber) As typeRangeVector  ' Range vector - relative to Bot center
                trvRangerBeams = ftnRangerBeamsRelativeBotCenter(tbpPossiblePose)
                Dim sngRMS_RangerError As Single = ftnRangerToMapRMS_Error(tbpPossiblePose, intRangerValidBeams, intNumberOfRangers, boolEndPointModel, trvRangerBeams) ' 
                If (ftnRMS_RangeError < sngOneDotFiveSigma) And (sngRMS_RangerError < sngOneDotFiveSigma) Then
                    ftnRMS_RangeError = Sqrt((ftnRMS_RangeError * intNumberOfBeams) ^ 2 + (sngRMS_RangerError * intNumberOfRangers) ^ 2) / (intNumberOfBeams + intNumberOfRangers)
                ElseIf (intRangerValidBeams >= 12) Then
                    ftnRMS_RangeError = sngRMS_RangerError
                End If
                intValidBeamCounter += intRangerValidBeams
            End If
        Catch ex As Exception
            ftnRMS_RangeError = sngOneDotFiveSigma
            frmDebug.subWriteLogs(Color.Green, " EXCEPTION in ftnRMS_RangeError")
        End Try
    End Function

    Function ftnRangerBeamsRelativeBotCenter(ByVal tbpEstimatedPose As typeBotPose) As typeRangeVector() ' Find Range vectors - relative to Bot center
        Dim trvRangerVector(intHighestEnabledRangerNumber) As typeRangeVector
        Try
            If boolRangersAreEnabled Then
                For intRangerNumber As Short = 1 To intHighestEnabledRangerNumber ' 
                    If trfRangeFinders(intRangerNumber).boolEnable Then _
                        trvRangerVector(intRangerNumber) = frmRangeFinder.ftnRangeFinderVector(tbpEstimatedPose, intRangerNumber, trfRangeFinders(intRangerNumber).sngRange, trfRangeFinders(intRangerNumber).intOrientation)
                Next intRangerNumber
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Green, " EXCEPTION in ftnRangerBeamsRelativeBotCenter")
        End Try
        Return trvRangerVector
    End Function




    Function ftnScanToMapRMS_Error(ByVal tbpPossiblePose As typeBotPose, _
                                     ByRef intValidBeamCounter As Short, ByVal boolEndPointModel As Boolean) As Single
        ' Find the correlation of a particular pose to the range data (higher =  better)
        ' boolEndPointModel = Don't check for intervening walls
        Dim intDistanceError, intDistanceErrorCumulativeSquared As Short
        Try
            If boolKinectConnected Then
                ftnScanToMapRMS_Error = 0
                Dim boolGreySpace As Boolean = False
                Dim trvVector As typeRangeVector
                intDistanceErrorCumulativeSquared = 0
                intValidBeamCounter = 0
                Dim tmcVectorOrigin As typeMapCoordinate = ftnLocationFromEstimatedPose(tbpPossiblePose, tmcScanVectorOffset)
                For intBeamNumber As Short = 0 To 127 Step 128 / intNumberOfBeams
                    trvVector.sngAngle = ftnConvertTo360(trvScanVectorsCopy(intBeamNumber).sngAngle + tbpPossiblePose.Orientation)
                    trvVector.sngRange = trvScanVectorsCopy(intBeamNumber).sngRange
                    If (trvVector.sngRange = 0) Then  ' 
                        intDistanceError = Round(sngWeakMatch)
                    ElseIf (trvVector.sngRange > ftnInchesToMapUnits(intScanVectorsMaxRangeInches)) Then
                        intDistanceError = Round(sngWeakMatch)
                    Else
                        intDistanceError = ftnRangeVectorVsWallError(tmcVectorOrigin, trvVector, boolEndPointModel, boolGreySpace)
                        If (intDistanceError >= sngOneDotFiveSigma) Then
                            If boolGreySpace Then intValidBeamCounter += 2 ' Grey zone = cluttered environment / may have false echos
                        Else
                            intValidBeamCounter += 3
                        End If
                    End If
                    intDistanceErrorCumulativeSquared += Min(intDistanceError * intDistanceError, 50)
                Next intBeamNumber
                intValidBeamCounter = intValidBeamCounter \ 3
                If (intValidBeamCounter > intNumberOfBeams \ 2) Then
                    ftnScanToMapRMS_Error = Min(Sqrt(intDistanceErrorCumulativeSquared) / intNumberOfBeams, sngOneDotFiveSigma)
                Else
                    ftnScanToMapRMS_Error = sngOneDotFiveSigma
                End If
            Else
                ftnScanToMapRMS_Error = sngOneDotFiveSigma
            End If
        Catch ex As Exception
            'frmDebug.subWriteLogs(Color.Green, " EXCEPTION in ftnScanToMapRMS_Error")
            frmDebug.subWriteLogs(Color.Green, " EXECEPTION in ftnScanToMapRMS_Error")
            frmDebug.subWriteLogs(Color.Green, " Pose =  " & tbpPossiblePose.Location.X & ",  " & tbpPossiblePose.Location.Y & " at " & tbpPossiblePose.Orientation & " degrees")
            frmDebug.subWriteLogs(Color.Green, " intValidBeamCounter = " & intValidBeamCounter)
            frmDebug.subWriteLogs(Color.Green, " intDistanceError = " & intDistanceError)
            frmDebug.subWriteLogs(Color.Green, " intDistanceErrorCumulativeSquared = " & intDistanceErrorCumulativeSquared)
            frmDebug.subWriteLogs(Color.Green, " ftnScanToMapRMS_Error = " & ftnScanToMapRMS_Error)
            ftnScanToMapRMS_Error = sngOneDotFiveSigma
        End Try
    End Function

    Function ftnRangerToMapRMS_Error(ByVal tbpEstimatedPose As typeBotPose, ByRef intValidBeamCounter As Short, ByRef intNumberOfRangers As Short, _
                                       ByVal boolEndPointModel As Boolean, ByVal trvRangerBeams() As typeRangeVector) As Single
        ' Find the correlation of a particular pose to the range data (higher =  better)
        ' boolEndPointModel = Don't check for intervening walls
        Try
            If boolRangersAreEnabled Then
                ftnRangerToMapRMS_Error = 0
                intNumberOfRangers = 0
                Dim boolGreySpace As Boolean = False
                Dim intDistanceError As Short
                Dim trvVector As typeRangeVector
                intValidBeamCounter = 0
                For intRangerNumber As Short = 1 To intHighestEnabledRangerNumber ' 
                    If trfRangeFinders(intRangerNumber).boolEnable Then
                        intNumberOfRangers += 1
                        trvVector.sngRange = trvRangerBeams(intRangerNumber).sngRange
                        trvVector.sngAngle = ftnConvertTo360(trvRangerBeams(intRangerNumber).sngAngle + tbpEstimatedPose.Orientation)
                        intDistanceError = Abs(ftnRangeVectorVsWallError(tbpEstimatedPose.Location, trvVector, boolEndPointModel, boolGreySpace))
                        ftnRangerToMapRMS_Error += Min(intDistanceError * intDistanceError, 50)
                        If boolGreySpace And (intDistanceError >= sngOneDotFiveSigma) Then
                            intValidBeamCounter += 2 ' Grey zone = cluttered environment / may have false echos
                        ElseIf (intDistanceError < sngOneDotFiveSigma) Then
                            intValidBeamCounter += 3
                        End If
                    End If
                Next intRangerNumber
                intValidBeamCounter = intValidBeamCounter \ 3
                If (intValidBeamCounter >= (intNumberOfRangers \ 2)) Then
                    ftnRangerToMapRMS_Error = Min(Sqrt(ftnRangerToMapRMS_Error) / intNumberOfRangers, sngOneDotFiveSigma)
                Else
                    ftnRangerToMapRMS_Error = sngOneDotFiveSigma
                End If
            Else
                ftnRangerToMapRMS_Error = sngOneDotFiveSigma
            End If
        Catch ex As Exception
            ftnRangerToMapRMS_Error = sngOneDotFiveSigma
            frmDebug.subWriteLogs(Color.Green, " EXCEPTION in ftnRangerToMapRMS_Error")
        End Try
    End Function

    Function ftnGreySpace(ByVal tmcLocation As typeMapCoordinate) As Boolean
        ' return true if location is in Grey zone = cluttered environment
        Dim byteCellContent As Byte = byteNavigationMap(tmcLocation.X, tmcLocation.Y)
        ftnGreySpace = (byteCellContent > 32) And (byteCellContent < 128)
    End Function

    Function ftnPoseUncertaintyFrom2DScan(ByVal tbpCurrentPose As typeBotPose, _
                                          ByVal sngRMS_RangeError As Single) As typeBotPose
        ' Reduce pose uncertainty if we have a good match        
        ftnPoseUncertaintyFrom2DScan = tbpCurrentPose
        If (sngRMS_RangeError <= sngExcellentMatch) Then
            ftnPoseUncertaintyFrom2DScan.LocationUncertainty.X = Min(ftnPoseUncertaintyFrom2DScan.LocationUncertainty.X, ftnInchesToMapUnits(10) / sngGridSize)
            ftnPoseUncertaintyFrom2DScan.LocationUncertainty.Y = Min(ftnPoseUncertaintyFrom2DScan.LocationUncertainty.Y, ftnPoseUncertaintyFrom2DScan.LocationUncertainty.X)
            ftnPoseUncertaintyFrom2DScan.OrientationUncertainty = Min(ftnPoseUncertaintyFrom2DScan.OrientationUncertainty, 10)
        ElseIf (sngRMS_RangeError <= sngGoodMatch) Then
            ftnPoseUncertaintyFrom2DScan.LocationUncertainty.X = Min(ftnPoseUncertaintyFrom2DScan.LocationUncertainty.X, ftnInchesToMapUnits(12) / sngGridSize)
            ftnPoseUncertaintyFrom2DScan.LocationUncertainty.Y = Min(ftnPoseUncertaintyFrom2DScan.LocationUncertainty.Y, ftnPoseUncertaintyFrom2DScan.LocationUncertainty.X)
            ftnPoseUncertaintyFrom2DScan.OrientationUncertainty = Min(ftnPoseUncertaintyFrom2DScan.OrientationUncertainty, 12)
        ElseIf (sngRMS_RangeError < sngWeakMatch) Then
            ftnPoseUncertaintyFrom2DScan.LocationUncertainty.X = Min(ftnPoseUncertaintyFrom2DScan.LocationUncertainty.X, ftnInchesToMapUnits(14) / sngGridSize)
            ftnPoseUncertaintyFrom2DScan.LocationUncertainty.Y = Min(ftnPoseUncertaintyFrom2DScan.LocationUncertainty.Y, ftnPoseUncertaintyFrom2DScan.LocationUncertainty.X)
            ftnPoseUncertaintyFrom2DScan.OrientationUncertainty = Min(ftnPoseUncertaintyFrom2DScan.OrientationUncertainty, 16)
        End If
        ftnPoseUncertaintyFrom2DScan = ftnMinPoseUncertainty(ftnPoseUncertaintyFrom2DScan)
    End Function



    '--------------------------------------  Non Flat Wall  ----------------------------------------------------------------------

    Function ftnNonFlatWall() As Boolean
        ' long flat walls which fill the Kinect's FOV can cause localization errors
        ftnNonFlatWall = False
        Try
            Dim sngKinectRange(128) As Single
            Dim sngKinectSlope(128) As Single
            Dim intBeamInterval As Short = 128 \ intNumberOfBeams
            Dim intTwoSamplesPerVector As Short = intBeamInterval + intBeamInterval
            For intBeamNumber As Short = 0 To 127 Step intBeamInterval
                sngKinectRange(intBeamNumber) = trvScanVectorsCopy(intBeamNumber).sngRange
            Next intBeamNumber
            If (sngKinectRange(intTwoSamplesPerVector) = 0) Then sngKinectRange(intTwoSamplesPerVector) = sngKinectRange(3 * intBeamInterval)
            If (sngKinectRange(intBeamInterval) = 0) Then sngKinectRange(intBeamInterval) = sngKinectRange(intTwoSamplesPerVector)
            If (sngKinectRange(0) = 0) Then sngKinectRange(0) = sngKinectRange(intBeamInterval)
            sngKinectSlope(intBeamInterval) = sngKinectRange(intBeamInterval) - sngKinectRange(0)
            Dim intI_MinusSamplesPerVector As Short
            For I As Short = intTwoSamplesPerVector To 127 Step intBeamInterval
                intI_MinusSamplesPerVector = I - intBeamInterval
                If (sngKinectRange(I) = 0) Then sngKinectRange(I) = sngKinectRange(intI_MinusSamplesPerVector) ' get rid of range = 0 entries
                sngKinectSlope(I) = sngKinectRange(I) - sngKinectRange(intI_MinusSamplesPerVector) ' calculate slopes
                If (sngKinectRange(I) <> 0) And (sngKinectRange(intI_MinusSamplesPerVector) <> 0) Then
                    If (Abs(sngKinectRange(I) - sngKinectRange(intI_MinusSamplesPerVector)) > sngTwoInches) Then ' Exit if Spike (Discontinuity)
                        ftnNonFlatWall = True
                        Exit Function
                    End If
                End If
            Next I
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Green, " EXCEPTION in ftnNonFlatWall")
        End Try
    End Function




End Module
