﻿'Various checks used to verify a localization result.
Imports System.Math

Module modLocalizeChecks

    Function ftnLocationValidForLM(ByVal tmcPossibleLocation As typeMapCoordinate, ByVal LandMark As typeLandMark, ByVal tbpReferencePose As typeBotPose) As Boolean
        tmcPossibleLocation = ftnCheckMapArrayLimits(tmcPossibleLocation)
        If ftnLocationWithinPose(tmcPossibleLocation, tbpReferencePose) And _
            Not ftnInWall(tmcPossibleLocation) And _
           ftnLocationWithinLMValidRegion(tmcPossibleLocation, LandMark) Then
            ftnLocationValidForLM = True
            If (Not boolBotIsMoving Or boolVerboseLogFile) Then subDrawSmallLimeCircle(tmcPossibleLocation) 'DEBUG DEBUG DEBUG
        Else
            ftnLocationValidForLM = False
            If (Not boolBotIsMoving Or boolVerboseLogFile) Then subDrawSmallRedCircle(tmcPossibleLocation) 'DEBUG DEBUG DEBUG
        End If
    End Function

    Function ftnLocationOKForLM(ByVal tmcPossibleLocation As typeMapCoordinate, ByVal tlmLandMark As typeLandMark, ByVal tbpReferencePose As typeBotPose) As Boolean
        tmcPossibleLocation = ftnCheckMapArrayLimits(tmcPossibleLocation)
        If ftnLocationWithinPose(tmcPossibleLocation, tbpReferencePose) And _
           Not boolOccupiedCell(tmcPossibleLocation.X, tmcPossibleLocation.Y) And _
           ftnLocationWithinLMValidRegion(tmcPossibleLocation, tlmLandMark) Then
            ftnLocationOKForLM = True
        Else
            ftnLocationOKForLM = False
        End If
    End Function



    Public Function ftnValidatePose(ByVal tbpCandidatePose As typeBotPose, ByVal tbpReferencePose As typeBotPose, ByRef boolValidLocation As Boolean) As typeBotPose
        If Not ftnMapLocationsEqual(ftnCheckMapArrayLimits(tbpCandidatePose.Location), tbpCandidatePose.Location) Then
            boolValidLocation = False
            ftnValidatePose = tbpReferencePose
        ElseIf ftnInWall(tbpCandidatePose.Location) Then
            boolValidLocation = False
            ftnValidatePose = tbpReferencePose
        ElseIf (Not ftnLocationWithinPose(tbpCandidatePose.Location, tbpReferencePose)) Then
            boolValidLocation = False
            ftnValidatePose = tbpReferencePose
        ElseIf Not ftnAnglesWithinUncertainty(tbpCandidatePose.Orientation, tbpReferencePose.Orientation, tbpReferencePose.OrientationUncertainty) Then
            boolValidLocation = False
            ftnValidatePose = tbpReferencePose
        Else
            boolValidLocation = True
            ftnValidatePose = tbpCandidatePose
        End If
    End Function

    Public Function ftnBotWithinLMValidRegion(ByVal LandMark As typeLandMark) As Boolean
        If (((tbpBotPose.Location.X + tbpBotPose.LocationUncertainty.X) >= LandMark.LandMarkValidRegion.X) And _
            ((tbpBotPose.Location.X - tbpBotPose.LocationUncertainty.X) <= (LandMark.LandMarkValidRegion.X + LandMark.LandMarkValidRegion.Width)) And _
            ((tbpBotPose.Location.Y + tbpBotPose.LocationUncertainty.Y) >= LandMark.LandMarkValidRegion.Y) And _
            ((tbpBotPose.Location.Y - tbpBotPose.LocationUncertainty.Y) <= (LandMark.LandMarkValidRegion.Y + LandMark.LandMarkValidRegion.Height))) Then
            ftnBotWithinLMValidRegion = True
        Else
            ftnBotWithinLMValidRegion = False
        End If
    End Function


    Public Function ftnLocationWithinPose(ByVal tmcPossibleLocation As typeMapCoordinate, ByVal tbpReferencePose As typeBotPose) As Boolean
        If (Abs(tmcPossibleLocation.X - tbpReferencePose.Location.X) <= tbpReferencePose.LocationUncertainty.X) And _
           (Abs(tmcPossibleLocation.Y - tbpReferencePose.Location.Y) <= tbpReferencePose.LocationUncertainty.Y) Then
            ftnLocationWithinPose = True
        Else
            ftnLocationWithinPose = False
        End If
    End Function


    Function ftnLocationWithinLMValidRegion(ByVal tmcPossibleLocation As typeMapCoordinate, ByVal LandMark As typeLandMark) As Boolean
        If ((tmcPossibleLocation.X >= LandMark.LandMarkValidRegion.X) And _
            (tmcPossibleLocation.X <= (LandMark.LandMarkValidRegion.X + LandMark.LandMarkValidRegion.Width)) And _
            (tmcPossibleLocation.Y >= LandMark.LandMarkValidRegion.Y) And _
            (tmcPossibleLocation.Y <= (LandMark.LandMarkValidRegion.Y + LandMark.LandMarkValidRegion.Height))) Then
            ftnLocationWithinLMValidRegion = True
        Else
            ftnLocationWithinLMValidRegion = False
        End If
    End Function

    Function ftnPoseWithinLMValidRegion(ByVal tmcPossiblePose As typeBotPose, ByVal LandMark As typeLandMark) As Boolean
        If (((tmcPossiblePose.Location.X + tmcPossiblePose.LocationUncertainty.X) >= LandMark.LandMarkValidRegion.X) And _
            ((tmcPossiblePose.Location.X - tmcPossiblePose.LocationUncertainty.X) <= (LandMark.LandMarkValidRegion.X + LandMark.LandMarkValidRegion.Width)) And _
            ((tmcPossiblePose.Location.Y + tmcPossiblePose.LocationUncertainty.Y) >= LandMark.LandMarkValidRegion.Y) And _
            ((tmcPossiblePose.Location.Y - tmcPossiblePose.LocationUncertainty.Y) <= (LandMark.LandMarkValidRegion.Y + LandMark.LandMarkValidRegion.Height))) Then
            ftnPoseWithinLMValidRegion = True
        Else
            ftnPoseWithinLMValidRegion = False
        End If
    End Function

    Public Function ftnBotToLMAngleCheck(ByVal LandMarkDetails As typeLandMark, ByVal sngBottoLMRelativeAngle As Single) As Boolean
        Dim sngBotToLMAngleAsMeasured, sngBotToLMAngleOnMap As Single

        sngBotToLMAngleOnMap = ftnConvertTo360(ftnBotToLMAngle(LandMarkDetails))
        sngBotToLMAngleAsMeasured = ftnConvertTo360(tbpBotPose.Orientation + sngBottoLMRelativeAngle)
        ftnBotToLMAngleCheck = ftnAnglesWithinUncertainty(sngBotToLMAngleOnMap, sngBotToLMAngleAsMeasured, tbpBotPose.OrientationUncertainty)
        If (ftnBotToLMAngleCheck = False) And boolVerboseLogFile Then
            frmDebug.subWriteLogs(Color.Violet, "    Angle to LM is wrong: ")
            frmDebug.subWriteLogs(Color.Violet, "          Measured Angle: " & sngBotToLMAngleAsMeasured & " degrees ")
            frmDebug.subWriteLogs(Color.Violet, "          Expected angle: " & sngBotToLMAngleOnMap & " +/- " & tbpBotPose.OrientationUncertainty & " degrees")
        End If
    End Function



    Public Function ftnBotToLMAngle(ByVal tlmLandMarkDetails As typeLandMark) As Single
        ftnBotToLMAngle = ftnP1ToP2Angle(tbpBotPose.Location, tlmLandMarkDetails.LandMarkLocation)
    End Function



    Public Function ftnInWall(ByVal tmcPossibleLocation As typeMapCoordinate) As Boolean
        ftnInWall = (byteNavigationMap(tmcPossibleLocation.X, tmcPossibleLocation.Y) >= byteNoGoNearWallSpaceThreshold)   'byteNoGoNearWallSpaceThreshold = 23,   constMaxBufferSpaceCost = 32
    End Function





End Module
