﻿Imports System.Math

Module modNavigationPath
    Dim sngTurnAngle As Single
    Const constNearWall = 512

    '=================================================

    '  This module is responsible for:
    '    - operates on map array generated by the modNavigationArray module
    '    - Generates a path (shortest distance) between the bot and the destination
    '

    '=================================================

    '  Convert the map array to a Path Matrix where the Destination is assigned the lowest value (ie Cost) 
    '  and the robot is assigned a much higher value.  The shortest path between robot and Destination may 
    ' be found by following the gradient.  i.e. Starting at the Robot's location, find the neighboring cell with the lowest value.
    ' This will be the next cell on the path to the destination.  Repeat until the destination is reached.


    Sub subGeneratePathMatrix()
        '  This routine generates a matrix representing the obstacles in the map
        '  byteNavigationMap:  0 = free,  255 = occupied
        ReDim intPathArray(CInt(GotABotClass.imgMap.Width), CInt(GotABotClass.imgMap.Height))
        strStatusDriving = intDrivingSequenceNumber & " : Generating Path Matrix "
        GotABotClass.subUpdateStatusLine()
        Dim intX, intY As Short
        Dim byteCellValue As Byte
        byteNoGoNearWallSpaceThreshold = (constMaxBufferSpaceCost - ftnMapUnitsToInches(sngBotRadius) - 0.5)  ' June2012 = 32-8.5-.5 = 23
        Dim byteNearWallBufferThreshold As Byte = constMaxBufferSpaceCost - intBufferSpaceMaxSize ' June2012 intBufferSpaceMaxSize =15
        If (byteNearWallBufferThreshold < constMinimumBufferSpaceCost) Then byteNearWallBufferThreshold = constMinimumBufferSpaceCost
        For intX = 0 To GotABotClass.imgMap.Width - 0
            For intY = 0 To GotABotClass.imgMap.Height - 0
                byteCellValue = byteNavigationMap(intX, intY)
                If (byteCellValue = 255) Then
                    intPathArray(intX, intY) = 65535
                ElseIf (byteCellValue > byteNoGoNearWallSpaceThreshold) And (byteCellValue < constMaxBufferSpaceCost) Then ' constMaxBufferSpaceCost = 32
                    intPathArray(intX, intY) = constMaxPathCost - 4096   'constNearWall    ' constNearWall = 512
                ElseIf (byteCellValue >= byteNearWallBufferThreshold) Then
                    intPathArray(intX, intY) = 4 + byteCellValue
                Else
                    intPathArray(intX, intY) = byteCellValue
                End If
            Next
        Next
    End Sub

    Dim boolCostIsUsed(constMaxPathCost) As Boolean  ' speed up ftnGeneratePathCosts by defining unused intNextLocationCost values
    Dim intMaxThreshold As UShort
    Function ftnGeneratePathCosts(ByVal tmcDestination As typeMapCoordinate, ByVal tmcStartLocation As typeMapCoordinate) As Boolean
        ' Returns true if a good path is found
        Try
            ftnGeneratePathCosts = True
            strStatusDriving = intDrivingSequenceNumber & " : Finding Path to Destination "
            GotABotClass.subUpdateStatusLine()
            Dim intX, intY As Short
            Dim intX_Max As Short = UBound(intPathArray, 1) - 1
            Dim intY_Max As Short = UBound(intPathArray, 2) - 1
            Dim intNextLocationCost As UShort = constDestinationCost + 1 ' the destination is the lowest cost. constDestinationCost = 4096
            intMaxThreshold = constNearWall
            If (ftnInWall(tbpBotPose.Location)) Then intMaxThreshold += 1

            ReDim boolCostIsUsed(constMaxPathCost) ' set all elements to False
            intPathArray(tmcDestination.X, tmcDestination.Y) = intNextLocationCost ' Start path at destination = lowest cost

            Dim timeOut As DateTime = Now().AddMilliseconds(30000)
            Dim intCurrentLocationCost As UShort = constDestinationCost
            While (intPathArray(tmcStartLocation.X, tmcStartLocation.Y) < constDestinationCost) And Not boolStop And (intNextLocationCost <> constMaxPathCost)
                For intX = 0 To intX_Max ' Work backward from the destination to the bot.  Each iteration (Location), add to the cost
                    For intY = 0 To intY_Max
                        If ((intPathArray(intX, intY) = intNextLocationCost)) Then
                            subUpdateCellNeighbors(intX, intY, intNextLocationCost + 1)
                            intCurrentLocationCost = intNextLocationCost + 1
                        End If
                    Next intY
                Next intX
                intNextLocationCost += 1
                While Not boolCostIsUsed(intNextLocationCost) And (intNextLocationCost <> constMaxPathCost)
                    intNextLocationCost += 1
                End While
                If boolSlam Then If (Now() > timeOut) Then Exit While ' structured this way for speed, in non Slam mode only boolSlam needs to be evaluated not Now()
            End While
            If (intNextLocationCost = constMaxPathCost) Or (boolSlam And (Now() > timeOut)) Then
                ftnGeneratePathCosts = False
                If Not boolSlam Then
                    strStatusDriving = intDrivingSequenceNumber & " : Invalid Destination. -"
                    GotABotClass.subUpdateAllText(tbpBotPose)
                    ftnSendCommand("play phrase 'Error - Invalid Destination'")
                End If
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Green, " EXCEPTION in ftnGeneratePathCosts")
        End Try
    End Function

    Sub subUpdateCellNeighbors(ByVal intX As Short, ByVal intY As Short, ByVal intNextLocationCost As UShort)
        '  Called from subroutine ftnGeneratePathCosts.  
        '  Up/Dn and Left/Right neighbors of the selected cell are incremented by +1.
        '  Diagonal Neighbors are incremented by +2
        Try
            Dim intCellX, intCellY As Short
            Dim intCellX_Max As Short = UBound(intPathArray, 1)
            Dim intCellY_Max As Short = UBound(intPathArray, 2)

            intNextLocationCost += 1  ' cost = +2 for Up/Dn and Left/Right neighbors
            intCellX = intX
            For intCellY = (intY - 1) To (intY + 1) Step 2 ' up and down neighbors
                If ((intCellX <> -1) And (intCellY <> -1) And intCellY <> intCellY_Max) Then _
                    subUpdateCell(intCellX, intCellY, intNextLocationCost)
            Next intCellY
            intCellY = intY
            For intCellX = (intX - 1) To (intX + 1) Step 2 ' left and right neighbors
                If ((intCellX <> -1) And intCellX <> intCellX_Max And (intCellY <> -1)) Then _
                    subUpdateCell(intCellX, intCellY, intNextLocationCost)
            Next intCellX

            intNextLocationCost += 1  ' cost = 2 for diagonal neighbors
            For intCellX = (intX - 1) To (intX + 1) Step 2 ' diagonals
                For intCellY = (intY - 1) To (intY + 1) Step 2 ' 
                    If ((intCellX <> -1) And intCellX <> intCellX_Max And (intCellY <> -1) And intCellY <> intCellY_Max) Then _
                      subUpdateCell(intCellX, intCellY, intNextLocationCost)
                Next intCellY
            Next intCellX
            'subDoEventsIfMainThread()
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Green, " EXCEPTION in subUpdateCellNeighbors")
        End Try
    End Sub


    Private Sub subUpdateCell(ByVal intCellX As Short, ByVal intCellY As Short, ByVal intNextLocationCost As UShort)
        Try
            If (intPathArray(intCellX, intCellY) <= intMaxThreshold) Then
                intPathArray(intCellX, intCellY) += intNextLocationCost
                boolCostIsUsed(intPathArray(intCellX, intCellY)) = True
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Green, " EXCEPTION in subUpdateCell")
        End Try
    End Sub

    Public Function ftnLowestValueNeighbor(ByVal tmcCurrentLocation As typeMapCoordinate) As typeMapCoordinate
        ' This function finds the next Location on the path as we move
        ' from bot to destination.  The next Location is the neighbor of tmcCurrentLocation
        ' which has the lowest intPathArray value).
        Try
            ftnLowestValueNeighbor = tmcCurrentLocation
            Dim tmcCandidateLocation As typeMapCoordinate
            Dim intLowestValue As UShort = intPathArray(ftnLowestValueNeighbor.X, ftnLowestValueNeighbor.Y)
            Dim intCandidateValue As UShort
            For tmcCandidateLocation.X = (tmcCurrentLocation.X - 1) To (tmcCurrentLocation.X + 1)
                For tmcCandidateLocation.Y = (tmcCurrentLocation.Y - 1) To (tmcCurrentLocation.Y + 1)
                    tmcCandidateLocation = ftnCheckMapArrayLimits(tmcCandidateLocation)
                    intCandidateValue = intPathArray(tmcCandidateLocation.X, tmcCandidateLocation.Y)
                    If ((intCandidateValue >= constDestinationCost) And (intCandidateValue < intLowestValue)) Then
                        ftnLowestValueNeighbor = tmcCandidateLocation
                        intLowestValue = intCandidateValue
                    End If
                Next tmcCandidateLocation.Y
            Next tmcCandidateLocation.X
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Green, " EXCEPTION in ftnLowestValueNeighbor")
        End Try
    End Function



    '==============================================================================================
    '   The following functions are used to find the intermediate steps along the path from a start location
    '    to the Destination
    '==============================================================================================

    Function ftnFindTargetPose(ByVal tbpEstimatedPose As typeBotPose) As typeBotPose
        ' This function searches for the furthest point on the path that it can safely reach using
        ' an optional 'turn' command followed by a move 'fwd' or 'MoveXY'command.  
        ftnFindTargetPose = tbpEstimatedPose
        Try
            Dim boolClearPath As Boolean = False
            Dim strCurrentTypeOfMove As String = "Fwd"
            Dim tbpFurthestSafePose As typeBotPose = tbpEstimatedPose
            Dim sngFurthustSafeDistance As Single
            Dim sngCurrentSafeDistance As Single = 0
            Do           'see if we can go furthur without hitting a wall
                boolClearPath = False
                sngCurrentSafeDistance = ftnDistanceBetweenLocations(tbpEstimatedPose.Location, ftnFindTargetPose.Location)
                If (sngCurrentSafeDistance > sngFurthustSafeDistance) Then
                    tbpFurthestSafePose = ftnFindTargetPose
                    sngFurthustSafeDistance = sngCurrentSafeDistance
                End If
                ftnFindTargetPose.Location = ftnNextLocationRequiringTurn(tbpEstimatedPose, ftnFindTargetPose.Location) 'Find the next Turn  
                tbpEstimatedPose.Orientation = ftnP1ToP2Angle(tbpEstimatedPose.Location, ftnFindTargetPose.Location) ' Orientation required to drive from tbpEstimatedPose.Location to ftnFindTargetPose
                If (strCurrentTypeOfMove = "Fwd") Then
                    boolClearPath = ftnClearPathFwd(tbpEstimatedPose, ftnFindTargetPose.Location) ' can we drive forward to this location without coming too close to a wall?
                    If Not boolClearPath And boolEnableXY_Motion Then
                        strCurrentTypeOfMove = "XY"
                        boolClearPath = ftnMoveXY_ConditionsMet(tbpEstimatedPose, ftnFindTargetPose.Location)
                    End If
                ElseIf boolEnableXY_Motion Then
                    boolClearPath = ftnMoveXY_ConditionsMet(tbpEstimatedPose, ftnFindTargetPose.Location)
                End If
            Loop Until (boolStop Or Not boolClearPath) _
                         Or ftnMapLocationsEqual(ftnFindTargetPose.Location, tmcDestination) '_
            If Not boolClearPath Then ftnFindTargetPose = tbpFurthestSafePose
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Green, " EXCEPTION in ftnFindTargetPose")
        End Try
    End Function


    Function ftnNextLocationRequiringTurn(ByVal tbpEstimatedPose As typeBotPose, ByVal tmcNextLocation As typeMapCoordinate) As typeMapCoordinate
        ' returns the first cell on the path which would requires a turn to reach
        Try
            ftnNextLocationRequiringTurn = tmcNextLocation
            Do
                ftnNextLocationRequiringTurn = ftnLowestValueNeighbor(ftnNextLocationRequiringTurn)
                subDoEventsIfMainThread()
            Loop Until boolStop _
                     Or ftnTurnGT3(tbpEstimatedPose, ftnNextLocationRequiringTurn) _
                     Or (intPathArray(ftnNextLocationRequiringTurn.X, ftnNextLocationRequiringTurn.Y) < constDestinationCost) _
                     Or ftnMapLocationsEqual(ftnNextLocationRequiringTurn, tmcDestination)
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Green, " EXCEPTION in ftnNextLocationRequiringTurn")
        End Try
    End Function

    Function ftnTypeOfMove(ByVal tbpEstimatedPose As typeBotPose, ByVal tbpTargetPose As typeBotPose) As String
        Try
            If ftnClearPathFwd(tbpEstimatedPose, tbpTargetPose.Location) Then
                ftnTypeOfMove = "Fwd"
            ElseIf ftnMoveXY_ConditionsMet(tbpEstimatedPose, tbpTargetPose.Location) Then
                ftnTypeOfMove = "XY"
            ElseIf ftnTurnGT3(tbpEstimatedPose, tbpTargetPose.Location) Then
                ftnTypeOfMove = "Turn"
            Else
                ftnTypeOfMove = "XY"
            End If
        Catch ex As Exception
            ftnTypeOfMove = "XY"
            frmDebug.subWriteLogs(Color.Green, " EXCEPTION in ftnTypeOfMove")
        End Try
    End Function

    Function ftnMoveXY_ConditionsMet(ByVal tbpEstimatedPose As typeBotPose, ByVal tmcNextLocation As typeMapCoordinate) As Boolean
        Try
            ftnMoveXY_ConditionsMet = False
            If boolEnableXY_Motion Then
                Dim intTurnAngle As Short = ftnTurnAngle(tbpEstimatedPose, tmcNextLocation)
                If (Abs(intTurnAngle) <= 80) Then
                    If (ftnDistanceBetweenLocations(tbpEstimatedPose.Location, tmcNextLocation) > sngTenInches) Then
                        ftnMoveXY_ConditionsMet = ftnClearPathXY(tbpEstimatedPose, tmcNextLocation)
                    End If
                End If
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Green, " EXCEPTION in ftnMoveXY_ConditionsMet")
        End Try
    End Function

    Function ftnTurnGT3(ByVal tbpEstimatedPose As typeBotPose, ByVal tmcNextLocation As typeMapCoordinate) As Boolean
        ftnTurnGT3 = (Abs(ftnTurnAngle(tbpEstimatedPose, tmcNextLocation)) > 3)
    End Function


    Function ftnClearPathFwd(ByVal tbpEstimatedPose As typeBotPose, ByVal tmcTargetLocation As typeMapCoordinate) As Boolean
        ' this function returns true if the path between tbpEstimatedPose.Location and tmcTargetLocation is clear of obstructions
        ' and not near a wall
        Try
            ftnClearPathFwd = True
            If ftnTurnGT3(tbpEstimatedPose, tmcTargetLocation) Then
                ftnClearPathFwd = False
            Else
                Dim sngDistanceToTarget As Single = ftnDistanceBetweenLocations(tbpEstimatedPose.Location, tmcTargetLocation)
                If (sngDistanceToTarget > sngFiveInches) Then
                    Dim sngDistanceFromStart As Single = sngFiveInches
                    Dim tmcCheckLocation As typeMapCoordinate
                    Do
                        tmcCheckLocation = ftnLocationFromEstimatedPose(tbpEstimatedPose, 0, sngDistanceFromStart)
                        ftnClearPathFwd = ftnSafeLocationForMove(tbpEstimatedPose, tmcCheckLocation, sngDistanceFromStart)
                        sngDistanceFromStart += sngOneInch
                    Loop Until ftnMapLocationsEqual(tmcCheckLocation, tmcTargetLocation) Or (sngDistanceFromStart > sngDistanceToTarget) Or boolStop Or Not ftnClearPathFwd
                End If
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Green, " EXCEPTION in ftnClearPathFwd")
        End Try
    End Function


    Function ftnClearPathXY(ByVal tbpEstimatedPose As typeBotPose, ByVal tmcTargetLocation As typeMapCoordinate) As Boolean
        ' this function returns true if the path between tbpEstimatedPose.Location and tmcTargetLocation is clear of obstructions
        ' and not near a wall
        Try
            ftnClearPathXY = True
            Dim sngDistanceToTarget As Single = (ftnDistanceBetweenLocations(tbpEstimatedPose.Location, tmcTargetLocation) * 1.5) ' Rough approximation
            If (sngDistanceToTarget > sngFiveInches) Then
                Dim tbpTargetPose As typeBotPose = tbpEstimatedPose
                tbpTargetPose.Location = tmcTargetLocation
                Dim sngDistanceFromStart As Single = sngFiveInches
                Dim tbpCheckPose As typeBotPose
                Do
                    tbpCheckPose = ftnXY_DR(tbpEstimatedPose, sngDistanceFromStart, tbpTargetPose, False)
                    ftnClearPathXY = ftnSafeLocationForMove(tbpEstimatedPose, tbpCheckPose.Location, sngDistanceFromStart)
                    sngDistanceFromStart += sngTwoInches
                Loop Until ftnMapLocationsEqual(tbpCheckPose.Location, tmcTargetLocation) Or (sngDistanceFromStart > sngDistanceToTarget) Or boolStop Or Not ftnClearPathXY
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Green, " EXCEPTION in ftnClearPathXY")
        End Try
    End Function



    Function ftnSafeLocationForMove(ByVal tbpStartPose As typeBotPose, ByVal tmcCheckLocation As typeMapCoordinate, ByVal sngDistanceFromStart As Single) As Boolean
        Try
            ftnSafeLocationForMove = True
            Dim byteCheckLocationValue As Byte = byteNavigationMap(tmcCheckLocation.X, tmcCheckLocation.Y)
            If (byteCheckLocationValue >= (byteNoGoNearWallSpaceThreshold - 5)) Then
                If (sngDistanceFromStart >= sngfortyInches) And (byteCheckLocationValue > (byteNoGoNearWallSpaceThreshold - 5)) Then
                    ftnSafeLocationForMove = False
                ElseIf (sngDistanceFromStart >= sngTenInches) And (byteCheckLocationValue > (byteNoGoNearWallSpaceThreshold - 3)) Then
                    ftnSafeLocationForMove = False
                ElseIf (byteCheckLocationValue >= (byteNoGoNearWallSpaceThreshold - 2)) _
                    And Not (byteCheckLocationValue <= byteNavigationMap(tbpStartPose.Location.X, tbpStartPose.Location.Y)) Then  ' July2014 byteNoGoNearWallSpaceThreshold = 23
                    ftnSafeLocationForMove = False
                End If
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Green, " EXCEPTION in ftnSafeLocationForMove")
        End Try
    End Function

End Module
